@@ -1,5 +1,5 @@
 /* assuan-buffer.c - read and send data
- *	Copyright (C) 2001, 2002, 2003, 2004 Free Software Foundation, Inc.
+ * Copyright (C) 2001, 2002, 2003, 2004, 2006 Free Software Foundation, Inc.
  *
  * This file is part of Assuan.
  *
@@ -14,8 +14,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA 
+ * License along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -30,8 +29,11 @@
 #endif
 #include "assuan-defs.h"
 
+
+/* Extended version of write(2) to guarantee that all bytes are
+   written.  Returns 0 on success or -1 and ERRNO on failure. */
 static int
-writen (ASSUAN_CONTEXT ctx, const char *buffer, size_t length)
+writen (assuan_context_t ctx, const char *buffer, size_t length)
 {
   while (length)
     {
@@ -49,9 +51,11 @@ writen (ASSUAN_CONTEXT ctx, const char *buffer, size_t length)
   return 0;  /* okay */
 }
 
-/* Read an entire line.  */
+/* Read an entire line. Returns 0 on success or -1 and ERRNo on
+   failure.  EOF is indictated by setting the integer at address
+   R_EOF.  */
 static int
-readline (ASSUAN_CONTEXT ctx, char *buf, size_t buflen,
+readline (assuan_context_t ctx, char *buf, size_t buflen,
 	  int *r_nread, int *r_eof)
 {
   size_t nleft = buflen;
@@ -88,8 +92,9 @@ readline (ASSUAN_CONTEXT ctx, char *buf, size_t buflen,
 }
 
 
-int
-_assuan_read_line (ASSUAN_CONTEXT ctx)
+/* Function returns an Assuan error. */
+assuan_error_t
+_assuan_read_line (assuan_context_t ctx)
 {
   char *line = ctx->inbound.line;
   int nread, atticlen;
@@ -97,7 +102,7 @@ _assuan_read_line (ASSUAN_CONTEXT ctx)
   char *endp = 0;
 
   if (ctx->inbound.eof)
-    return -1;
+    return _assuan_error (-1);
 
   atticlen = ctx->inbound.attic.linelen;
   if (atticlen)
@@ -128,19 +133,20 @@ _assuan_read_line (ASSUAN_CONTEXT ctx)
   if (rc)
     {
       if (ctx->log_fp)
-	fprintf (ctx->log_fp, "%s[%u.%p] DBG: <- [Error: %s]\n",
+	fprintf (ctx->log_fp, "%s[%u.%d] DBG: <- [Error: %s]\n",
 		 assuan_get_assuan_log_prefix (),
-                 (unsigned int)getpid (), ctx, strerror (errno));
-      return ASSUAN_Read_Error;
+                 (unsigned int)getpid (), (int)ctx->inbound.fd,
+                 strerror (errno));
+      return _assuan_error (ASSUAN_Read_Error);
     }
   if (!nread)
     {
       assert (ctx->inbound.eof);
       if (ctx->log_fp)
-	fprintf (ctx->log_fp, "%s[%u.%p] DBG: <- [EOF]\n",
+	fprintf (ctx->log_fp, "%s[%u.%d] DBG: <- [EOF]\n",
 		 assuan_get_assuan_log_prefix (),
-                 (unsigned int)getpid (), ctx);
-      return -1;
+                 (unsigned int)getpid (), (int)ctx->inbound.fd);
+      return _assuan_error (-1);
     }
 
   ctx->inbound.attic.pending = 0;
@@ -151,7 +157,9 @@ _assuan_read_line (ASSUAN_CONTEXT ctx)
 
   if (endp)
     {
+      unsigned monitor_result;
       int n = endp - line + 1;
+
       if (n < nread)
 	/* LINE contains more than one line.  We copy it to the attic
 	   now as handlers are allowed to modify the passed
@@ -168,11 +176,20 @@ _assuan_read_line (ASSUAN_CONTEXT ctx)
       *endp = 0;
 
       ctx->inbound.linelen = endp - line;
-      if (ctx->log_fp)
+
+      monitor_result = (ctx->io_monitor
+                        ? ctx->io_monitor (ctx, 0,
+                                           ctx->inbound.line,
+                                           ctx->inbound.linelen)
+                        : 0);
+      if ( (monitor_result & 2) )
+        ctx->inbound.linelen = 0;
+      
+      if (ctx->log_fp && !(monitor_result & 1))
 	{
-	  fprintf (ctx->log_fp, "%s[%u.%p] DBG: <- ",
+	  fprintf (ctx->log_fp, "%s[%u.%d] DBG: <- ",
 		   assuan_get_assuan_log_prefix (),
-                   (unsigned int)getpid (), ctx);
+                   (unsigned int)getpid (), (int)ctx->inbound.fd);
 	  if (ctx->confidential)
 	    fputs ("[Confidential data not shown]", ctx->log_fp);
 	  else
@@ -186,13 +203,14 @@ _assuan_read_line (ASSUAN_CONTEXT ctx)
   else
     {
       if (ctx->log_fp)
-	fprintf (ctx->log_fp, "%s[%u.%p] DBG: <- [Invalid line]\n",
+	fprintf (ctx->log_fp, "%s[%u.%d] DBG: <- [Invalid line]\n",
 		 assuan_get_assuan_log_prefix (),
-                 (unsigned int)getpid (), ctx);
+                 (unsigned int)getpid (), (int)ctx->inbound.fd);
       *line = 0;
       ctx->inbound.linelen = 0;
-      return ctx->inbound.eof ? ASSUAN_Line_Not_Terminated
-	: ASSUAN_Line_Too_Long;
+      return _assuan_error (ctx->inbound.eof 
+                            ? ASSUAN_Line_Not_Terminated
+                            : ASSUAN_Line_Too_Long);
     }
 }
 
@@ -207,12 +225,12 @@ _assuan_read_line (ASSUAN_CONTEXT ctx)
    See also: assuan_pending_line().
 */
 assuan_error_t
-assuan_read_line (ASSUAN_CONTEXT ctx, char **line, size_t *linelen)
+assuan_read_line (assuan_context_t ctx, char **line, size_t *linelen)
 {
   assuan_error_t err;
 
   if (!ctx)
-    return ASSUAN_Invalid_Value;
+    return _assuan_error (ASSUAN_Invalid_Value);
 
   err = _assuan_read_line (ctx);
   *line = ctx->inbound.line;
@@ -224,7 +242,7 @@ assuan_read_line (ASSUAN_CONTEXT ctx, char **line, size_t *linelen)
 /* Return true if a full line is buffered (i.e. an entire line may be
    read without any I/O).  */
 int
-assuan_pending_line (ASSUAN_CONTEXT ctx)
+assuan_pending_line (assuan_context_t ctx)
 {
   return ctx && ctx->inbound.attic.pending;
 }
@@ -234,52 +252,61 @@ assuan_error_t
 _assuan_write_line (assuan_context_t ctx, const char *prefix,
                     const char *line, size_t len)
 {
-  int rc = 0;
+  assuan_error_t rc = 0;
   size_t prefixlen = prefix? strlen (prefix):0;
+  unsigned int monitor_result;
 
   /* Make sure that the line is short enough. */
   if (len + prefixlen + 2 > ASSUAN_LINELENGTH)
     {
       if (ctx->log_fp)
-        fprintf (ctx->log_fp, "%s[%u.%p] DBG: -> "
+        fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> "
                  "[supplied line too long -truncated]\n",
                  assuan_get_assuan_log_prefix (),
-                 (unsigned int)getpid (), ctx);
+                 (unsigned int)getpid (), (int)ctx->inbound.fd);
       if (prefixlen > 5)
         prefixlen = 5;
       if (len > ASSUAN_LINELENGTH - prefixlen - 2)
         len = ASSUAN_LINELENGTH - prefixlen - 2 - 1;
     }
 
+  monitor_result = (ctx->io_monitor
+                    ? ctx->io_monitor (ctx, 1, line, len)
+                    : 0);
+
   /* Fixme: we should do some kind of line buffering.  */
-  if (ctx->log_fp)
+  if (ctx->log_fp && !(monitor_result & 1))
     {
-      fprintf (ctx->log_fp, "%s[%u.%p] DBG: -> ",
+      fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> ",
 	       assuan_get_assuan_log_prefix (),
-               (unsigned int)getpid (), ctx);
+               (unsigned int)getpid (), (int)ctx->inbound.fd);
       if (ctx->confidential)
 	fputs ("[Confidential data not shown]", ctx->log_fp);
       else
-	_assuan_log_print_buffer (ctx->log_fp, line, len);
+        {
+          if (prefixlen)
+            _assuan_log_print_buffer (ctx->log_fp, prefix, prefixlen);
+          _assuan_log_print_buffer (ctx->log_fp, line, len);
+        }
       putc ('\n', ctx->log_fp);
     }
 
-  if (prefixlen)
+  if (prefixlen && !(monitor_result & 2))
     {
       rc = writen (ctx, prefix, prefixlen);
       if (rc)
-        rc = ASSUAN_Write_Error;
+        rc = _assuan_error (ASSUAN_Write_Error);
     }
-  if (!rc)
+  if (!rc && !(monitor_result & 2))
     {
       rc = writen (ctx, line, len);
       if (rc)
-        rc = ASSUAN_Write_Error;
+        rc = _assuan_error (ASSUAN_Write_Error);
       if (!rc)
         {
           rc = writen (ctx, "\n", 1);
           if (rc)
-            rc = ASSUAN_Write_Error;
+            rc = _assuan_error (ASSUAN_Write_Error);
         }
     }
   return rc;
@@ -287,13 +314,13 @@ _assuan_write_line (assuan_context_t ctx, const char *prefix,
 
 
 assuan_error_t 
-assuan_write_line (ASSUAN_CONTEXT ctx, const char *line)
+assuan_write_line (assuan_context_t ctx, const char *line)
 {
   size_t len;
   const char *s;
 
   if (!ctx)
-    return ASSUAN_Invalid_Value;
+    return _assuan_error (ASSUAN_Invalid_Value);
 
   /* Make sure that we never take a LF from the user - this might
      violate the protocol. */
@@ -301,10 +328,10 @@ assuan_write_line (ASSUAN_CONTEXT ctx, const char *line)
   len = s? (s-line) : strlen (line);
 
   if (ctx->log_fp && s)
-    fprintf (ctx->log_fp, "%s[%u.%p] DBG: -> "
-             "[supplied line contained a LF -truncated]\n",
+    fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> "
+             "[supplied line contained a LF - truncated]\n",
              assuan_get_assuan_log_prefix (),
-             (unsigned int)getpid (), ctx);
+             (unsigned int)getpid (), (int)ctx->inbound.fd);
 
   return _assuan_write_line (ctx, NULL, line, len);
 }
@@ -312,11 +339,11 @@ assuan_write_line (ASSUAN_CONTEXT ctx, const char *line)
 
 
 /* Write out the data in buffer as datalines with line wrapping and
-   percent escaping.  This function is used for GNU's custom streams */
+   percent escaping.  This function is used for GNU's custom streams. */
 int
 _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
 {
-  ASSUAN_CONTEXT ctx = cookie;
+  assuan_context_t ctx = cookie;
   size_t size = orig_size;
   char *line;
   size_t linelen;
@@ -329,7 +356,9 @@ _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
   line += linelen;
   while (size)
     {
-      /* insert data line header */
+      unsigned int monitor_result;
+
+      /* Insert data line header. */
       if (!linelen)
         {
           *line++ = 'D';
@@ -337,7 +366,7 @@ _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
           linelen += 2;
         }
       
-      /* copy data, keep some space for the CRLF and to escape one character */
+      /* Copy data, keep space for the CRLF and to escape one character. */
       while (size && linelen < LINELENGTH-2-2)
         {
           if (*buffer == '%' || *buffer == '\r' || *buffer == '\n')
@@ -355,13 +384,19 @@ _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
           size--;
         }
       
+      
+      monitor_result = (ctx->io_monitor
+                        ? ctx->io_monitor (ctx, 1,
+                                           ctx->outbound.data.line, linelen)
+                        : 0);
+
       if (linelen >= LINELENGTH-2-2)
         {
-          if (ctx->log_fp)
+          if (ctx->log_fp && !(monitor_result & 1))
             {
-	      fprintf (ctx->log_fp, "%s[%u.%p] DBG: -> ",
+	      fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> ",
 		       assuan_get_assuan_log_prefix (),
-                       (unsigned int)getpid (), ctx);
+                       (unsigned int)getpid (), (int)ctx->inbound.fd);
 
               if (ctx->confidential)
                 fputs ("[Confidential data not shown]", ctx->log_fp);
@@ -373,9 +408,10 @@ _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
             }
           *line++ = '\n';
           linelen++;
-          if (writen (ctx, ctx->outbound.data.line, linelen))
+          if ( !(monitor_result & 2)
+               && writen (ctx, ctx->outbound.data.line, linelen))
             {
-              ctx->outbound.data.error = ASSUAN_Write_Error;
+              ctx->outbound.data.error = _assuan_error (ASSUAN_Write_Error);
               return 0;
             }
           line = ctx->outbound.data.line;
@@ -393,9 +429,10 @@ _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
 int
 _assuan_cookie_write_flush (void *cookie)
 {
-  ASSUAN_CONTEXT ctx = cookie;
+  assuan_context_t ctx = cookie;
   char *line;
   size_t linelen;
+  unsigned int monitor_result;
 
   if (ctx->outbound.data.error)
     return 0;
@@ -403,13 +440,19 @@ _assuan_cookie_write_flush (void *cookie)
   line = ctx->outbound.data.line;
   linelen = ctx->outbound.data.linelen;
   line += linelen;
+
+  monitor_result = (ctx->io_monitor
+                    ? ctx->io_monitor (ctx, 1,
+                                       ctx->outbound.data.line, linelen)
+                    : 0);
+  
   if (linelen)
     {
-      if (ctx->log_fp)
+      if (ctx->log_fp && !(monitor_result & 1))
 	{
-	  fprintf (ctx->log_fp, "%s[%u.%p] DBG: -> ",
+	  fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> ",
 		   assuan_get_assuan_log_prefix (),
-                   (unsigned int)getpid (), ctx);
+                   (unsigned int)getpid (), (int)ctx->inbound.fd);
 	  if (ctx->confidential)
 	    fputs ("[Confidential data not shown]", ctx->log_fp);
 	  else
@@ -419,9 +462,10 @@ _assuan_cookie_write_flush (void *cookie)
 	}
       *line++ = '\n';
       linelen++;
-      if (writen (ctx, ctx->outbound.data.line, linelen))
+      if ( !(monitor_result & 2)
+           && writen (ctx, ctx->outbound.data.line, linelen))
         {
-          ctx->outbound.data.error = ASSUAN_Write_Error;
+          ctx->outbound.data.error = _assuan_error (ASSUAN_Write_Error);
           return 0;
         }
       ctx->outbound.data.linelen = 0;
@@ -449,12 +493,12 @@ _assuan_cookie_write_flush (void *cookie)
  **/
 
 assuan_error_t
-assuan_send_data (ASSUAN_CONTEXT ctx, const void *buffer, size_t length)
+assuan_send_data (assuan_context_t ctx, const void *buffer, size_t length)
 {
   if (!ctx)
-    return ASSUAN_Invalid_Value;
+    return _assuan_error (ASSUAN_Invalid_Value);
   if (!buffer && length)
-    return ASSUAN_Invalid_Value;
+    return _assuan_error (ASSUAN_Invalid_Value);
 
   if (!buffer)
     { /* flush what we have */
@@ -475,8 +519,17 @@ assuan_send_data (ASSUAN_CONTEXT ctx, const void *buffer, size_t length)
 }
 
 assuan_error_t
-assuan_sendfd (ASSUAN_CONTEXT ctx, int fd)
+assuan_sendfd (assuan_context_t ctx, assuan_fd_t fd)
 {
+  /* It is explicitly allowed to use (NULL, -1) as a runtime test to
+     check whether descriptor passing is available. */
+  if (!ctx && fd == ASSUAN_INVALID_FD)
+#ifdef USE_DESCRIPTOR_PASSING
+    return 0;
+#else
+    return _assuan_error (ASSUAN_Not_Implemented);
+#endif
+
   if (! ctx->io->sendfd)
     return set_error (ctx, Not_Implemented,
 		      "server does not support sending and receiving "
@@ -485,7 +538,7 @@ assuan_sendfd (ASSUAN_CONTEXT ctx, int fd)
 }
 
 assuan_error_t
-assuan_receivefd (ASSUAN_CONTEXT ctx, int *fd)
+assuan_receivefd (assuan_context_t ctx, assuan_fd_t *fd)
 {
   if (! ctx->io->receivefd)
     return set_error (ctx, Not_Implemented,
